@@ -34,6 +34,8 @@
  * 29 Mar 05, daved@physiol.usyd.edu.au: changes requested by ZT Smith
  * 16 Dec 07, daved@physiol.usyd.edu.au: fixed fore/background_begin error
  *                       and updated to GPL v3
+ * 09 Nov 08, arkadiusz.firus@gmail.com: adopt safe_printf
+ * 07 Oct 11, jf@dockes.org, fix to add_to_collection()
  *--------------------------------------------------------------------*/
 
 #ifdef HAVE_CONFIG_H
@@ -52,6 +54,10 @@
 #include <stdlib.h>
 #endif
 
+#ifndef HAVE_ATDARG_H
+#include <stdarg.h>
+#endif
+
 #include "malloc.h"
 #include "defs.h"
 #include "error.h"
@@ -82,8 +88,8 @@ typedef struct _stack {
 	struct _stack *next;
 } AttrStack;
 
-static AttrStack *stack_of_stacks = NULL;
-static AttrStack *stack_of_stacks_top = NULL;
+/*@null@*/ static AttrStack *stack_of_stacks = NULL;
+/*@null@*/ static AttrStack *stack_of_stacks_top = NULL;
 
 
 
@@ -100,10 +106,10 @@ attr_express_begin (int attr, char* param) {
 	switch(attr) 
 	{
 	case ATTR_BOLD: 
-		printf(op->bold_begin); 
+		if (safe_printf(0, op->bold_begin)) fprintf(stderr, TOO_MANY_ARGS, "bold_begin");;
 		break;
 	case ATTR_ITALIC: 
-		printf(op->italic_begin); 
+		if (safe_printf(0, op->italic_begin)) fprintf(stderr, TOO_MANY_ARGS, "italic_begin");;
 		break;
 
 	/* Various underlines, they all resolve to HTML's <u> */
@@ -115,11 +121,11 @@ attr_express_begin (int attr, char* param) {
 	case ATTR_2DOT_DASH_UL:
 	case ATTR_WORD_UL: 
 	case ATTR_UNDERLINE: 
-		printf(op->underline_begin); 
+		if (safe_printf(0, op->underline_begin)) fprintf(stderr, TOO_MANY_ARGS, "underline_begin");;
 		break;
 
 	case ATTR_DOUBLE_UL: 
-		printf(op->dbl_underline_begin); 
+		if (safe_printf(0, op->dbl_underline_begin)) fprintf(stderr, TOO_MANY_ARGS, "dbl_underline_begin");;
 		break;
 
 	case ATTR_FONTSIZE: 
@@ -127,48 +133,48 @@ attr_express_begin (int attr, char* param) {
 		break;
 
 	case ATTR_FONTFACE: 
-		printf(op->font_begin,param); 
+		if (safe_printf(1, op->font_begin,param)) fprintf(stderr, TOO_MANY_ARGS, "font_begin");;
 		break;
 
 	case ATTR_FOREGROUND: 
-		printf(op->foreground_begin, param); 
+		if (safe_printf(1, op->foreground_begin, param)) fprintf(stderr, TOO_MANY_ARGS, "foreground_begin");;
 		break;
 
 	case ATTR_BACKGROUND: 
 		if (!simple_mode)
-		  printf(op->background_begin,param); 
+		  if (safe_printf(1, op->background_begin,param)) fprintf(stderr, TOO_MANY_ARGS, "background_begin");;
 		break;
 
 	case ATTR_SUPER: 
-		printf(op->superscript_begin); 
+		if (safe_printf(0, op->superscript_begin)) fprintf(stderr, TOO_MANY_ARGS, "superscript_begin");;
 		break;
 	case ATTR_SUB: 
-		printf(op->subscript_begin); 
+		if (safe_printf(0, op->subscript_begin)) fprintf(stderr, TOO_MANY_ARGS, "subscript_begin");;
 		break;
 
 	case ATTR_STRIKE: 
-		printf(op->strikethru_begin); 
+		if (safe_printf(0, op->strikethru_begin)) fprintf(stderr, TOO_MANY_ARGS, "strikethru_begin");;
 		break;
 
 	case ATTR_DBL_STRIKE: 
-		printf(op->dbl_strikethru_begin); 
+		if (safe_printf(0, op->dbl_strikethru_begin)) fprintf(stderr, TOO_MANY_ARGS, "dbl_strikethru_begin");;
 		break;
 
 	case ATTR_EXPAND: 
-		printf(op->expand_begin, param); 
+		if (safe_printf(1, op->expand_begin, param)) fprintf(stderr, TOO_MANY_ARGS, "expand_begin");;
 		break;
 
 	case ATTR_OUTLINE: 
-		printf(op->outline_begin); 
+		if (safe_printf(0, op->outline_begin)) fprintf(stderr, TOO_MANY_ARGS, "outline_begin");;
 		break;
 	case ATTR_SHADOW: 
-		printf(op->shadow_begin); 
+		if (safe_printf(0, op->shadow_begin)) fprintf(stderr, TOO_MANY_ARGS, "shadow_begin");;
 		break;
 	case ATTR_EMBOSS: 
-		printf(op->emboss_begin); 
+		if (safe_printf(0, op->emboss_begin)) fprintf(stderr, TOO_MANY_ARGS, "emboss_begin");;
 		break;
 	case ATTR_ENGRAVE: 
-		printf(op->engrave_begin); 
+		if (safe_printf(0, op->engrave_begin)) fprintf(stderr, TOO_MANY_ARGS, "engrave_begin");;
 		break;
 
 	case ATTR_CAPS:
@@ -181,7 +187,7 @@ attr_express_begin (int attr, char* param) {
 			simulate_smallcaps = TRUE;
 		else {
 			if (op->small_caps_begin)
-				printf(op->small_caps_begin); 
+				if (safe_printf(0, op->small_caps_begin)) fprintf(stderr, TOO_MANY_ARGS, "small_caps_begin");;
 		}
 		break;
 	}
@@ -201,10 +207,10 @@ attr_express_end (int attr, char *param)
 	switch(attr) 
 	{
 	case ATTR_BOLD: 
-		printf(op->bold_end); 
+		if (safe_printf(0, op->bold_end)) fprintf(stderr, TOO_MANY_ARGS, "bold_end");;
 		break;
 	case ATTR_ITALIC: 
-		printf(op->italic_end); 
+		if (safe_printf(0, op->italic_end)) fprintf(stderr, TOO_MANY_ARGS, "italic_end");;
 		break;
 
 	/* Various underlines, they all resolve to HTML's </u> */
@@ -216,11 +222,11 @@ attr_express_end (int attr, char *param)
 	case ATTR_2DOT_DASH_UL: 
 	case ATTR_WORD_UL: 
 	case ATTR_UNDERLINE: 
-		printf(op->underline_end); 
+		if (safe_printf(0, op->underline_end)) fprintf(stderr, TOO_MANY_ARGS, "underline_end");;
 		break;
 
 	case ATTR_DOUBLE_UL: 
-		printf(op->dbl_underline_end); 
+		if (safe_printf(0, op->dbl_underline_end)) fprintf(stderr, TOO_MANY_ARGS, "dbl_underline_end");;
 		break;
 
 	case ATTR_FONTSIZE: 
@@ -228,47 +234,47 @@ attr_express_end (int attr, char *param)
 		break;
 
 	case ATTR_FONTFACE: 
-		printf(op->font_end); 
+		if (safe_printf(0, op->font_end)) fprintf(stderr, TOO_MANY_ARGS, "font_end");;
 		break;
 
 	case ATTR_FOREGROUND: 
-		printf(op->foreground_end); 
+		if (safe_printf(0, op->foreground_end)) fprintf(stderr, TOO_MANY_ARGS, "foreground_end");;
 		break;
 	case ATTR_BACKGROUND: 
 		if (!simple_mode)
-		  printf(op->background_end);
+		  if (safe_printf(0, op->background_end)) fprintf(stderr, TOO_MANY_ARGS, "background_end");;
 		break;
 
 	case ATTR_SUPER: 
-		printf(op->superscript_end); 
+		if (safe_printf(0, op->superscript_end)) fprintf(stderr, TOO_MANY_ARGS, "superscript_end");;
 		break;
 	case ATTR_SUB: 
-		printf(op->subscript_end); 
+		if (safe_printf(0, op->subscript_end)) fprintf(stderr, TOO_MANY_ARGS, "subscript_end");;
 		break;
 
 	case ATTR_STRIKE: 
-		printf(op->strikethru_end); 
+		if (safe_printf(0, op->strikethru_end)) fprintf(stderr, TOO_MANY_ARGS, "strikethru_end");;
 		break;
 
 	case ATTR_DBL_STRIKE: 
-		printf(op->dbl_strikethru_end); 
+		if (safe_printf(0, op->dbl_strikethru_end)) fprintf(stderr, TOO_MANY_ARGS, "dbl_strikethru_end");;
 		break;
 
 	case ATTR_OUTLINE: 
-		printf(op->outline_end); 
+		if (safe_printf(0, op->outline_end)) fprintf(stderr, TOO_MANY_ARGS, "outline_end");;
 		break;
 	case ATTR_SHADOW: 
-		printf(op->shadow_end); 
+		if (safe_printf(0, op->shadow_end)) fprintf(stderr, TOO_MANY_ARGS, "shadow_end");;
 		break;
 	case ATTR_EMBOSS: 
-		printf(op->emboss_end); 
+		if (safe_printf(0, op->emboss_end)) fprintf(stderr, TOO_MANY_ARGS, "emboss_end");;
 		break;
 	case ATTR_ENGRAVE: 
-		printf(op->engrave_end); 
+		if (safe_printf(0, op->engrave_end)) fprintf(stderr, TOO_MANY_ARGS, "engrave_end");;
 		break;
 
 	case ATTR_EXPAND: 
-		printf(op->expand_end); 
+		if (safe_printf(0, op->expand_end)) fprintf(stderr, TOO_MANY_ARGS, "expand_end");;
 		break;
 
 	case ATTR_CAPS:
@@ -281,7 +287,7 @@ attr_express_end (int attr, char *param)
 			simulate_smallcaps = FALSE;
 		else {
 			if (op->small_caps_end)
-				printf(op->small_caps_end); 
+				if (safe_printf(0, op->small_caps_end)) fprintf(stderr, TOO_MANY_ARGS, "small_caps_end");;
 		}
 		break;
 	}
@@ -325,7 +331,6 @@ attr_push(int attr, char* param)
 	attr_express_begin(attr, param);
 }
 
-#if 1 /* daved 0.20.2 */
 
 /*========================================================================
  * Name:	attr_get_param
@@ -359,8 +364,6 @@ attr_get_param(int attr)
 	return NULL;
 }
 
-#endif
-
 
 /*========================================================================
  * Name:	attrstack_copy_all
@@ -450,12 +453,6 @@ attrstack_push ()
 	}
 	stack_of_stacks_top = new_stack;
 	new_stack->tos = -1;
-
-	if (prev_stack) {
-		attrstack_unexpress_all (prev_stack);
-		attrstack_copy_all (prev_stack, new_stack);
-		attrstack_express_all ();
-	}
 }
 
 
@@ -566,8 +563,8 @@ attrstack_drop ()
 	}
 
 	attr_pop_all ();
-
 	prev_stack = stack_of_stacks;
+
 	while(prev_stack && prev_stack->next && prev_stack->next != stack)
 		prev_stack = prev_stack->next;
 
@@ -578,9 +575,8 @@ attrstack_drop ()
 		stack_of_stacks_top = NULL;
 		stack_of_stacks = NULL;
 	}
-	my_free ((void*) stack);
 
-	attrstack_express_all ();
+	my_free ((void*) stack);
 }
 
 /*========================================================================
@@ -667,3 +663,202 @@ attr_pop_dump() {
 	}
 }
 
+/*========================================================================
+ * Name:	safe_printf
+ * Purpose:	Prevents format string attack and writes empty string
+		instead of NULL.
+ * Args:	Number of parameters (without a string), string to write,
+		additional parameters to print (have to be strings).
+ * Returns:	Returns 0 if number of not escaped '%' in string
+		 is not greater than nr, else returns -1
+ *=======================================================================*/
+
+
+int
+safe_printf(int nr, char *string, ...)
+{
+
+	char *s;
+	int i = 0, ret_code = 0;
+	va_list arguments;
+
+	if (string == NULL)
+		printf("");
+	else
+	{
+		va_start(arguments, string);
+
+		for (; nr > 0; nr--)
+		{
+			while (string[i] != '\0' && (string[i] != '%' || (string[i] == '%' && (i != 0 && string[i-1] == '\\'))))
+			{
+				if (string[i] != '\\' || string[i+1] != '%')
+					printf("%c", string[i]);
+				i++;
+			}
+
+			if (string[i] != '\0')
+			{
+				s = va_arg(arguments, char *);
+				printf("%s", s);
+				i++;
+			}
+		}
+		va_end(arguments);
+
+		while (string[i] != '\0')
+		{
+			if (string[i] != '\\' || (string[i] == '\\' && string[i+1] != '%'))
+			{
+				if (string[i] != '%' || (string[i] == '%' && (i != 0 && string[i-1] == '\\')))
+					printf("%c", string[i]);
+				else
+					ret_code = -1;
+			}
+			i++;
+		}
+	}
+
+	return ret_code;
+}
+
+/*========================================================================
+ * Name:	assemble_string
+ * Purpose:	See Returns
+ * Args:	String to return and int to put into first parameter.
+ * Returns:	Returns first parameter where first not escaped
+		character % is substituted with second parameter.
+ *=======================================================================*/
+
+char *
+assemble_string(char *string, int nr)
+{
+
+	char *s, tmp[12];/* Number of characters that can be in int type (including '\0') - AF */
+	int i = 0, j = 0;
+
+	if (string == NULL)
+		return NULL;
+	else {
+		s = my_malloc(strlen(string) + 1 + 12/* Number of characters that can be in int type (including '\0') - AF */);
+		while(string[i] != '\0' && (string[i] != '%' || (string[i] == '%' && (i != 0 && string[i-1] == '\\')))) {
+			if (string[i] != '\\' || string[i+1] != '%') {
+				 s[j] = string[i];
+				j++;
+			}
+			i++;
+		}
+
+		if (string[i] != '\0') {
+			sprintf(tmp, "%d", nr);
+			strcpy(&s[j], tmp);
+			j = j + strlen(tmp);
+		}
+
+		while (string[i] != '\0') {
+			if (string[i] != '\\' || (string[i] == '\\' && string[i+1] != '%')) {
+				if (string[i] != '%' || (string[i] == '%' && (i != 0 && string[i-1] == '\\'))) {
+					 s[j] = string[i];
+					j++;
+				}
+				else {
+					/* More than one char % occured */			
+				}
+			}
+			i++;
+		}
+	}
+
+	s[j] = '\0';
+
+	return s;
+}
+
+
+/*========================================================================
+ * Name:	add_to_collection
+ * Purpose:	Adds (substitutes) element under index nr.
+ * Args:	Collection, index, text to add.
+ * Returns:	Collection.
+ *=======================================================================*/
+
+Collection *
+add_to_collection(Collection *col, int nr, char *text)
+{
+	Collection *c = col;
+
+	if (col == NULL)
+	{
+		col = (Collection *)my_malloc(sizeof(Collection));
+		col->nr = nr;
+		col->text = text;
+		col->next = NULL;
+	}
+	else
+	{
+            do	{
+			if (c->nr == nr)
+			{
+/* Here is a memory leak but not heavy. Do we need to care about this?
+				my_free(a->alias.text);
+*/
+				c->text = text;
+
+				return col;
+			}
+
+                        if (c->next == NULL)
+                            break;
+			c = c->next;
+            } while (1);
+
+		c->next = (Collection *)my_malloc(sizeof(Collection));
+		c->next->nr = nr;
+		c->next->text = text;
+		c->next->next = NULL;
+	}
+
+	return col;
+}
+
+/*========================================================================
+ * Name:	get_from_collection
+ * Purpose:	Search for element under index nr.
+ * Args:	Collection, index.
+ * Returns:	NULL or found elemet.
+ *=======================================================================*/
+
+char *
+get_from_collection(Collection *c, int nr)
+{
+	while (c != NULL)
+	{
+		if (c->nr == nr)
+			return c->text;
+
+		c = c->next;
+	}
+
+	return NULL;
+}
+
+/*========================================================================
+ * Name:	free_collection
+ * Purpose:	Release memory used by collection.
+ * Args:	Needless collection.
+ * Returns:	
+ *=======================================================================*/
+
+void
+free_collection(Collection *c)
+{
+	Collection *c2;
+
+	while (c != NULL)
+	{
+		c2 = c->next;
+		my_free((void *)c);
+		c = c2;
+	}
+}
+
